@@ -13,6 +13,9 @@ DNS::Unbound - libunbound in Perl
 
     my $dns = DNS::Unbound->new()->set_option( verbosity => 2 );
 
+    # This appears to be safe:
+    $dns->enable_threads();
+
     my $verbosity = $dns->get_option( 'verbosity' );
 
     $dns->set_option( verbosity => 1 + $verbosity );
@@ -60,7 +63,7 @@ use DNS::Unbound::X ();
 our ($VERSION);
 
 BEGIN {
-    $VERSION = '0.10_01';
+    $VERSION = '0.11_2';
     XSLoader::load();
 }
 
@@ -148,17 +151,9 @@ Runs a synchronous query for a given $NAME and $TYPE. $TYPE may be
 expressed numerically or, for convenience, as a string. $CLASS is
 optional and defaults to 1 (C<IN>), which is probably what you want.
 
-Returns a reference to a hash that corresponds
-to a libunbound C<struct ub_result>
-(cf. L<libunbound(3)|https://nlnetlabs.nl/documentation/unbound/libunbound/>),
-excluding C<len>, C<answer_packet>, and C<answer_len>.
+Returns a L<DNS::Unbound::Result> instance.
 
-B<NOTE:> Members of C<data> are in their DNS-native RDATA encodings.
-(libunbound doesn’t track which record type uses which encoding, so
-neither does DNS::Unbound.)
-To decode some common record types, see L</CONVENIENCE FUNCTIONS> below.
-
-Also B<NOTE:> libunbound doesn’t seem to offer effective controls for
+B<NOTE:> libunbound doesn’t seem to offer effective controls for
 timing out a synchronous query.
 If timeouts are relevant for you, you probably need
 to use C<resolve_async()> instead.
@@ -256,12 +251,14 @@ sub resolve_async {
 
     my $query_id = $async_ar->[1];
 
-    $dns{'id'} = $query_id;
+    $_[0]->{'_queries_dns'}{ $query_id } = \%dns;
+    $_[0]->{'_queries_lookup'}{ $query_id } = undef;
+
+    # NB: If %dns referenced $query it would be a circular reference.
+    @dns{'id', 'queries_lookup'} = ($query_id, $_[0]->{'_queries_lookup'});
 
     $query->_set_dns(\%dns);
 
-    $_[0]->{'_queries_hr'}{ $query_id } = $query;
-
     return $query;
 }
 
@@ -286,8 +283,8 @@ already been sent.
 
 Returns I<OBJ>.
 
-B<NOTE:> Perl’s relationship with threads is … complicated.
-This option is not well-tested. If in doubt, just skip it.
+B<NOTE:> Despite Perl’s iffy relationship with threads, this appears
+to work without issue.
 
 =cut
 
@@ -413,7 +410,8 @@ Gives the libunbound version string.
 
 =head1 METHODS FOR DEALING WITH ASYNCHRONOUS QUERIES
 
-The following methods correspond to their equivalents in libunbound.
+Unless otherwise noted, the following methods correspond to their
+equivalents in libunbound.
 
 =head2 I<OBJ>->poll()
 
@@ -463,17 +461,28 @@ sub process {
     return $ret;
 }
 
+=head2 I<OBJ>->count_pending_promises()
+
+Returns the number of outstanding asynchronous queries.
+
+=cut
+
+sub count_pending_promises {
+    my ($self) = @_;
+
+    return $self->{'_queries_lookup'} ? 0 + keys %{ $self->{'_queries_lookup'} } : 0;
+}
+
 #----------------------------------------------------------------------
 
 sub _check_promises {
     my ($self) = @_;
 
-    if ( my $asyncs_hr = $self->{'_queries_hr'} ) {
-        for (values %$asyncs_hr) {
-            my $dns_hr = $_->_get_dns();
-
+    if ( my $asyncs_hr = $self->{'_queries_dns'} ) {
+        for my $dns_hr (values %$asyncs_hr) {
             if (defined $dns_hr->{'value'}) {
                 delete $asyncs_hr->{ $dns_hr->{'id'} };
+                delete $self->{'_queries_lookup'}{ $dns_hr->{'id'} };
 
                 my $key;
 
@@ -519,7 +528,7 @@ functions (but not as class methods). In addition to these,
 L<Socket> provides the C<inet_ntoa()> and C<inet_ntop()>
 functions for decoding the values of C<A> and C<AAAA> records.
 
-B<NOTE:> L<DNS::Unbound::Result>’s C<to_net_dns()> provides a heavier but
+B<NOTE:> L<DNS::Unbound::Result>’s C<to_net_dns_rrs()> provides a heavier but
 more robust way to parse query result data.
 
 =head2 $decoded = decode_name($encoded)
@@ -555,8 +564,7 @@ sub decode_character_strings {
 sub DESTROY {
     $_[0]->{'_destroyed'} ||= $_[0]->{'_ub'} && do {
         if ($$ == $_[0]->{'_pid'}) {
-            if (my $queries_hr = $_[0]->{'_queries_hr'}) {
-                $_->_forget_dns() for values %$queries_hr;
+            if (my $queries_hr = $_[0]->{'_queries_dns'}) {
                 %$queries_hr = ();
             }
         }
